from flask.json import JSONEncoder
from datetime import datetime
from bson import ObjectId, Timestamp
from .uorm.models.base_model import BaseModel
from .uorm.db import ObjectsCursor
import simplejson


def default(o):  # pylint: disable=method-hidden

    if isinstance(o, ObjectId):
        return str(o)
    if isinstance(o, Timestamp):
        return o.time
    if isinstance(o, (ObjectsCursor, set)):
        return list(o)
    if isinstance(o, BaseModel):
        return o.to_dict()
    if isinstance(o, datetime):
        return o.isoformat()
    raise TypeError(f"{repr(o)} is not JSON serializable")


def dumps(o, default=default, **kwargs):
    return simplejson.dumps(o, default=default, **kwargs)


def loads(text: str, **kwargs):
    return simplejson.loads(text, **kwargs)



from pymongo import ASCENDING, DESCENDING
from unittest import TestCase
from ..uorm.models.fields import StringField
from ..uorm.models.base_model import BaseModel


class TestBaseModel(TestCase):
    def test_empty(self):
        model = BaseModel()
        self.assertIsNone(model._id)
        self.assertCountEqual(model.__fields__, ["_id"])

    def test_not_empty(self):
        class Model(BaseModel):
            field: StringField()

        model = Model({"field": "value"})
        self.assertEqual(model.field, "value")

    def test_collection_name(self):
        class Model(BaseModel):
            COLLECTION = "model"

        model = BaseModel()
        self.assertEqual(model.collection, "base_model")
        model = Model()
        self.assertEqual(model.collection, "model")

    def test_fields(self):
        class Model(BaseModel):
            string_f: StringField()

        model = Model()
        self.assertCountEqual(model.__fields__, ["_id", "string_f"])

    def test_rejected_fields(self):
        class Model(BaseModel):
            string_f: StringField()
            string_rf: StringField(rejected=True)

        model = Model()
        self.assertCountEqual(model.__fields__, ["_id", "string_f", "string_rf"])
        self.assertCountEqual(model.__rejected_fields__, ["string_rf"])

    def test_restricted_fields(self):
        class Model(BaseModel):
            string_f: StringField()
            string_rf: StringField(restricted=True)

        model = Model()
        self.assertCountEqual(model.__fields__, ["_id", "string_f", "string_rf"])
        self.assertCountEqual(model.__restricted_fields__, ["string_rf"])

    def test_indexes(self):
        class Model(BaseModel):
            string_f: StringField()
            string_if: StringField(index=True)
            string_uf: StringField(unique=True)
            string_duf: StringField(index=DESCENDING, unique=True)

        model = Model()

        self.assertCountEqual(
            model.__def_indexes__,
            [
                [[("string_if", ASCENDING)], {}],
                [[("string_uf", ASCENDING)], {"unique": True}],
                [[("string_duf", DESCENDING)], {"unique": True}],
            ],
        )

    def test_to_dict(self):
        class Model(BaseModel):
            field1: StringField()
            field2: StringField()
            field3: StringField(restricted=True)

        model = Model({"field1": "value1", "field2": "value2", "field3": "value3",})

        self.assertDictEqual(
            model.to_dict(), {"_id": None, "field1": "value1", "field2": "value2"}
        )

        self.assertDictEqual(
            model.to_dict(fields=["field1", "field2", "field3"]),
            {"field1": "value1", "field2": "value2"},
        )

        self.assertDictEqual(
            model.to_dict(include_restricted=True),
            {"_id": None, "field1": "value1", "field2": "value2", "field3": "value3"},
        )

        self.assertDictEqual(
            model.to_dict(fields=["field1", "field3"], include_restricted=True),
            {"field1": "value1", "field3": "value3"},
        )

        self.assertDictEqual(
            model.to_dict(
                fields=["field1", "field3", "bizzare"], include_restricted=True
            ),
            {"field1": "value1", "field3": "value3"},
        )




import logging


class ContextError(Exception):
    pass


_NOT_SET = object()


class _LineProfilerFuncs:
    """This is pseudo-LineProfiler that collects functions to profile later"""

    def __init__(self):
        self.functions = set()

    def __call__(self, func):
        self.functions.add(func)
        return func


def gen_ctx_prop(name, default=_NOT_SET):
    attr_name = f"_{name}"

    def get_no_default(self):
        try:
            prop = getattr(self, attr_name)
        except AttributeError:
            raise ContextError(f"Attempted to use uninitialised {name}")
        return prop

    def get_default(self):
        return getattr(self, attr_name, default)

    def setter(self, value):
        try:
            getattr(self, attr_name)
        except AttributeError:
            setattr(self, attr_name, value)
            return
        # raise ContextError(f"Attempted to overwrite already initialised {name}")

    def deleter(self):
        delattr(self, attr_name)

    getter = get_no_default if default is _NOT_SET else get_default

    return property(fget=getter, fset=setter, fdel=deleter)


class _Context:
    # Although it is possible to assign generated properties to correct attributes
    # inside of _gen_prop(), doing so here explicitly declares them. It helps
    # linters and IDE auto-completers see the attributes and also:
    # >>> import this
    # Explicit is better than implicit
    envtype = gen_ctx_prop("envtype")
    cfg = gen_ctx_prop("cfg")
    log = gen_ctx_prop("log", default=logging)
    db = gen_ctx_prop("db")
    cache = gen_ctx_prop("cache")
    req_cache = gen_ctx_prop("req_cache")
    queue = gen_ctx_prop("queue")
    line_profiler = _LineProfilerFuncs()


ctx = _Context()



from glasskit.base import Base


class App(Base):

    def setup_controllers(self):
        pass

    def after_setup(self):
        pass


def force_init_app():
    # touch Proxy object
    _ = app.__doc__


app = App()



import cProfile
import mtprof
import io
import pstats
import functools
import warnings

try:
    import line_profiler
except ImportError:
    warnings.warn("Could not import line_profiler. Line profiling is disabled")
    line_profiler = None

from . import ctx

from flask import g
from datetime import datetime


def before_request():
    from .api import get_boolean_request_param

    if get_boolean_request_param("profile"):
        if line_profiler and ctx.line_profiler.functions:
            g.line_profiler = line_profiler.LineProfiler(*ctx.line_profiler.functions)
            g.line_profiler.enable()
        if get_boolean_request_param("profile_threads"):
            g.profiler = mtprof.Profile()
        else:
            g.profiler = cProfile.Profile()
        g.profiler.enable()


def after_request(response):
    if hasattr(g, "profiler"):
        g.profiler.disable()
        strio = io.StringIO()
        ps = pstats.Stats(g.profiler, stream=strio).sort_stats("cumulative")
        ps.print_stats()
        ctx.log.debug(strio.getvalue())

    if hasattr(g, "line_profiler"):
        g.line_profiler.disable()
        strio = io.StringIO()
        g.line_profiler.print_stats(stream=strio)
        ctx.log.debug(strio.getvalue())

    return response


def error_log_timings(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        t1 = datetime.now()
        result = func(*args, **kwargs)
        t2 = datetime.now()
        ctx.log.error(
            "%s finished in %.3f seconds", func.__name__, (t2 - t1).total_seconds()
        )
        return result

    return wrapper



from .test_base_model import TestBaseModel
from .test_storable_model import TestStorableModel
from .test_sharded_model import TestShardedModel
from .test_submodel import TestStorableSubmodel, TestShardedSubmodel
from .test_validators import TestValidators



import argparse
import os
import re
import jinja2
import venv
import shutil
from subprocess import Popen
from glasskit.uorm.models.meta import snake_case


PROJECT_NAME_EXPRESSION = re.compile(r"^[a-zA-Z0-9_]+$")
FIELD_TYPE_MAP = {
    "any": "Field",
    "str": "StringField",
    "string": "StringField",
    "int": "IntField",
    "bool": "BoolField",
    "boolean": "BoolField",
    "float": "FloatField",
    "list": "ListField",
    "dict": "DictField",
    "oid": "ObjectIdField",
    "object_id": "ObjectIdField",
    "datetime": "DatetimeField",
    "date": "DatetimeField",
    "ts": "DatetimeField",
}


def find_project_name():
    project_name = None
    for dirname in os.listdir("."):
        if os.path.isdir(dirname):
            contents = list(os.listdir(dirname))
            if "controllers" in contents:
                project_name = dirname
                break
    return project_name


def copy_sources(src_dir, dst_dir, appname):
    context = {"appname": appname}
    for srcname in os.listdir(src_dir):

        srcname = os.path.join(src_dir, srcname)
        if srcname.endswith("__pycache__") or srcname.endswith(".pyc"):
            continue

        dstname = os.path.join(dst_dir, os.path.basename(srcname))
        dstname = dstname.replace("__appname__", appname)

        if os.path.isdir(srcname):
            os.makedirs(dstname)
            copy_sources(srcname, dstname, appname)
        elif os.path.isfile(srcname):
            if srcname.endswith(".tmpl"):
                dstname = dstname[:-5]
                with open(srcname) as tmplfile:
                    tmpl = jinja2.Template(tmplfile.read())
                    with open(dstname, "w") as outfile:
                        outfile.write(tmpl.render(context))
            else:
                shutil.copy(srcname, dstname)


def glasskit_directory():
    return os.path.abspath(os.path.dirname(__file__))


def run_create(args):
    appname = args.project_name[0]
    dirname = appname
    if not PROJECT_NAME_EXPRESSION.match(dirname):
        raise ValueError(f"Invalid project name {dirname}")
    dirname = os.path.abspath(f"./{dirname}")

    project_templates_dir = os.path.join(glasskit_directory(), "project_templates")

    os.makedirs(dirname, exist_ok=False)
    copy_sources(project_templates_dir, dirname, appname)

    print("creating virtualenv")
    curr_dir = os.path.abspath(os.curdir)
    try:
        os.chdir(dirname)
        venv.create(args.venv, with_pip=True)
        python_path = os.path.join(os.curdir, "venv/bin/pip3")
        p = Popen([python_path, "install", "-r", "requirements.txt"])
        p.wait()
        os.chmod("glass.py", 0o755)
    finally:
        os.chdir(curr_dir)


def run_new_model(args):
    project_name = find_project_name()
    if project_name is None:
        raise RuntimeError(
            "project name is not detected. are you inside your project folder?"
        )

    class_name = args.model_name[0]
    model_filename = snake_case(class_name) + ".py"
    model_filename = os.path.join(project_name, f"models/{model_filename}")

    fields = []
    for field_desc in args.fields:
        field_desc = field_desc.split(":")
        field_type = "string"
        if len(field_desc) > 1:
            field_type = field_desc[1]
        field_desc = field_desc[0]

        if field_type not in FIELD_TYPE_MAP:
            raise TypeError(f'invalid field type "{field_type}"')
        fields.append((field_desc, FIELD_TYPE_MAP[field_type]))
    model_imports = ", ".join({fd[1] for fd in fields})
    tmplfilename = os.path.join(glasskit_directory(), "unit_templates/model.py.tmpl")
    with open(tmplfilename) as tmplfile:
        tmpl = jinja2.Template(tmplfile.read())

    os.makedirs(os.path.dirname(model_filename), exist_ok=True)
    with open(model_filename, "w") as modelf:
        modelf.write(
            tmpl.render(
                {
                    "classname": class_name,
                    "model_imports": model_imports,
                    "fields": fields,
                }
            )
        )
    print(f"File {model_filename} created")


def run_new_controller(args):
    project_name = find_project_name()
    if project_name is None:
        raise RuntimeError(
            "project name is not detected. are you inside your project folder?"
        )

    cls_tokens = args.ctrl_class.split(".")
    cls_module_path = ".".join(cls_tokens[:-1])
    cls_name = cls_tokens[-1]

    path_tokens = args.ctrl_path[0].split("/")
    ctrl_path = "/".join(path_tokens[:-1])
    ctrl_name = path_tokens[-1]

    ctrl_filename = os.path.join(
        project_name, "controllers", ctrl_path, f"{ctrl_name}.py"
    )

    context = {
        "cls_module_path": cls_module_path,
        "cls_name": cls_name,
        "ctrl_name": ctrl_name,
        "require_auth": not args.disable_auth,
    }

    tmplfilename = os.path.join(
        glasskit_directory(), "unit_templates/controller.py.tmpl"
    )
    with open(tmplfilename) as tmplfile:
        tmpl = jinja2.Template(tmplfile.read())

    os.makedirs(os.path.dirname(ctrl_filename), exist_ok=True)
    with open(ctrl_filename, "w") as ctrlfile:
        ctrlfile.write(tmpl.render(**context))

    print(f"File {ctrl_filename} created")


def run_new_task(args):
    project_name = find_project_name()
    if project_name is None:
        raise RuntimeError(
            "project name is not detected. are you inside your project folder?"
        )

    class_name = args.class_name[0]
    fields = args.fields
    create_args_list = ", ".join(["cls"] + fields)

    tmplfilename = os.path.join(glasskit_directory(), "unit_templates/task.py.tmpl")
    with open(tmplfilename) as tmplfile:
        tmpl = jinja2.Template(tmplfile.read())

    task_type = snake_case(class_name).upper()

    context = {
        "task_type": task_type,
        "cls_name": class_name,
        "fields": fields,
        "create_args_list": create_args_list,
    }

    task_filename = snake_case(class_name).lower() + ".py"
    task_filename = os.path.join(project_name, "tasks", task_filename)

    with open(task_filename, "w") as outfile:
        outfile.write(tmpl.render(**context))

    print(f"File {task_filename} created")


def main():
    parser = argparse.ArgumentParser(prog="glasskit")
    subparsers = parser.add_subparsers(
        dest="command", required=True, help="sub-command help"
    )

    parser_create = subparsers.add_parser(
        "create", help="bootstrap a new glasskit project"
    )
    parser_create.add_argument(
        "project_name", nargs=1, type=str, help="project name to create"
    )
    parser_create.add_argument(
        "--venv", "-v", type=str, default="venv", help="venv directory name"
    )

    parser_new = subparsers.add_parser(
        "new", help="create a new unit, i.e. model, controller, task"
    )
    parser_new_subparsers = parser_new.add_subparsers(
        dest="unit", required=True, help="unit-type help"
    )
    new_model_parser = parser_new_subparsers.add_parser(
        "model", help="create a new model"
    )

    new_model_parser.add_argument("model_name", nargs=1, type=str, help="model_name")
    new_model_parser.add_argument(
        "fields",
        nargs="*",
        type=str,
        help="field descriptors: field1:any field2 "
        "field3:string "
        "field4:list etc...",
    )

    new_task_parser = parser_new_subparsers.add_parser("task", help="create a new task")
    new_task_parser.add_argument(
        "class_name", nargs=1, type=str, help="name of the task class"
    )
    new_task_parser.add_argument("fields", nargs="*", type=str, help="task fields list")

    new_controller_parser = parser_new_subparsers.add_parser(
        "controller", help="create a new controller"
    )
    new_controller_parser.add_argument(
        "ctrl_path", nargs=1, type=str, help="controller prefix path, i.e. api/v1/users"
    )
    new_controller_parser.add_argument(
        "--disable-auth", "-d", action="store_true", help="disable auth requirement"
    )
    new_controller_parser.add_argument(
        "--class",
        "-c",
        dest="ctrl_class",
        type=str,
        default="glasskit.controller.Controller",
        help="override default controller class",
    )
    args = parser.parse_args()

    if args.command == "create":
        run_create(args)
    elif args.command == "new":
        unit_type = args.unit
        if unit_type == "model":
            run_new_model(args)
        elif unit_type == "controller":
            run_new_controller(args)
        elif unit_type == "task":
            run_new_task(args)
        else:
            raise ValueError(f"unknown unit type f{unit_type}")


if __name__ == "__main__":
    main()



from .base_queue import BaseQueue


class DummyQueue(BaseQueue):
    """
    DummyQueue is for testing purposes only
    """

    def __init__(self, qcfg):
        super(DummyQueue, self).__init__(qcfg)
        self.queue = []

    def _enqueue(self, task):
        self.queue.append(task)

    def ack(self, task_id):
        return task_id

    def subscribe(self):
        pass

    def list_active_channels(self):
        return [{"chan": "dummy:local"}]

    @property
    def tasks(self):
        while len(self.queue):
            item = self.queue.pop(0)
            yield item



from traceback import format_exc
from werkzeug.exceptions import HTTPException

from . import ctx
from .api import json_response


class ApiError(Exception):

    status_code = 400
    error_key = "api_error"

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code:
            self.status_code = status_code
        self.payload = payload or {}

    def to_dict(self):
        data = self.payload
        data["error_key"] = self.error_key
        data["error"] = self.message
        return data

    def __repr__(self):
        return "%s: %s, status_code=%s" % (
            self.__class__.__name__,
            self.message,
            self.status_code,
        )

    def __str__(self):
        return "%s, status_code=%s" % (self.message, self.status_code)


class AuthenticationError(ApiError):

    status_code = 401
    error_key = "auth_error"
    auth_url = None

    def __init__(self, message="you must be authenticated first", payload=None):
        ApiError.__init__(self, message, payload=payload)
        if self.auth_url is None:
            oauth_cfg = ctx.cfg.get("oauth")
            if oauth_cfg:
                client_id = oauth_cfg.get("id")
                auth_url = oauth_cfg.get("authorize_url")
                callback_url = oauth_cfg.get("callback_url")
                if client_id and auth_url and callback_url:
                    AuthenticationError.auth_url = (
                        f"{auth_url}?response_type=code&"
                        f"client_id={client_id}&scope=user_info&redirect_uri={callback_url}"
                    )

    def to_dict(self):
        data = super().to_dict()
        data["oauth"] = self.auth_url
        if "state" not in data:
            data["state"] = "logged out"
        return data


class ConfigurationError(SystemExit):
    pass


class Forbidden(ApiError):
    error_key = "forbidden"
    status_code = 403


class IntegrityError(ApiError):
    error_key = "integrity_error"
    status_code = 409


class NotFound(ApiError):
    error_key = "not_found"
    status_code = 404


class InvalidShardId(ApiError):
    error_key = "intrnl_error"
    status_code = 500


class ShardIsReadOnly(IntegrityError):
    pass


class ModelDestroyed(IntegrityError):
    pass


class MissingSubmodel(IntegrityError):
    pass


class WrongSubmodel(IntegrityError):
    pass


class UnknownSubmodel(IntegrityError):
    pass


class InputDataError(ApiError):
    error_key = "bad_input"
    pass


class InvalidFieldType(ApiError):
    error_key = "bad_input_type"
    pass


def handle_api_error(error):
    return json_response(error.to_dict(), error.status_code)


def handle_other_errors(error):
    code = 400
    if hasattr(error, "code"):
        code = error.code
    if not (100 <= code < 600):
        code = 400
    if not issubclass(error.__class__, HTTPException):
        ctx.log.error(format_exc())
    return json_response({"error": str(error)}, code)



from . import Command
import os


class Shell(Command):

    DESCRIPTION = 'Run shell (using IPython if available)'

    def run(self):
        if os.path.isfile(".shellrc.py"):
            with open(".shellrc.py") as rcf:
                try:
                    exec(rcf.read())
                except:
                    from traceback import print_exc
                    print("Error running .shellrc.py script!")
                    print_exc()
        try:
            # trying IPython if installed...
            from IPython import embed
            embed(using=False)
        except ImportError:
            # ... or python default console if not
            try:
                # optional readline interface for history if installed
                import readline  # pylint: disable=possibly-unused-variable
            except ImportError:
                pass
            import code
            variables = globals().copy()
            variables.update(locals())
            shell = code.InteractiveConsole(variables)
            shell.interact()



from typing import Dict, Any, Union, List
from functools import partial
from pymongo.cursor import Cursor
from .storable_model import StorableModel
from ..errors import MissingShardId
from glasskit.errors import NotFound
from ..db import db, Shard, ObjectsCursor
from ..utils import resolve_id, ObjectId


class ShardedModel(StorableModel):

    def __init__(self, attrs: Dict[str, Any], shard_id: str = None):
        super().__init__(attrs)
        self._shard_id = shard_id
        if not self.is_new and self._shard_id is None:
            from traceback import print_stack
            print_stack()
            raise MissingShardId(
                "ShardedModel from database with missing shard_id - this must be a bug")

    @classmethod
    def create(cls, shard_id=None, **attrs) -> 'ShardedModel':
        return cls(attrs, shard_id)

    @property
    def _db(self) -> Shard:
        return db.shards[self._shard_id]

    def save(self, skip_callback=False, invalidate_cache=True) -> None:
        if self._shard_id is None:
            raise MissingShardId(
                "ShardedModel must have shard_id set before save")
        super().save(skip_callback, invalidate_cache)

    def _refetch_from_db(self) -> 'ShardedModel':
        return self.find_one(self._shard_id, {"_id": self._id})

    @classmethod
    def _get_possible_databases(cls):
        return list(db.shards.values())

    @classmethod
    def find(cls, shard_id: str, query: Dict = None, **kwargs) -> ObjectsCursor:
        if not query:
            query = {}
        return db.get_shard(shard_id).get_objs(
            cls._ctor,
            cls.collection,
            cls._preprocess_query(query),
            **kwargs
        )

    @classmethod
    def aggregate(cls, shard_id: str, pipeline: List, query: Dict = None, **kwargs) -> Cursor:
        if not query:
            query = {}
        pipeline = [{"$match": cls._preprocess_query(query)}] + pipeline
        return db.get_shard(shard_id).get_aggregated(cls.collection, pipeline, **kwargs)

    @classmethod
    def find_projected(cls, shard_id: str, query: Dict = None, projection=('_id',), **kwargs) -> Cursor:
        if not query:
            query = {}
        return db.get_shard(shard_id).get_objs_projected(
            cls.collection,
            cls._preprocess_query(query),
            projection=projection,
            **kwargs
        )

    @classmethod
    def find_one(cls, shard_id: str, query: Dict, **kwargs) -> 'ShardedModel':
        return db.get_shard(shard_id).get_obj(
            cls._ctor,
            cls.collection,
            cls._preprocess_query(query),
            **kwargs
        )

    @classmethod
    def get(cls, shard_id: str, expression: Any, raise_if_none=None) -> Union['ShardedModel', None]:
        if expression is None:
            return None
        expression = resolve_id(expression)
        if isinstance(expression, ObjectId):
            query = {"_id": expression}
        else:
            expression = str(expression)
            query = {cls.KEY_FIELD: expression}
        res = cls.find_one(shard_id, query)
        if res is None and raise_if_none is not None:
            if isinstance(raise_if_none, Exception):
                raise raise_if_none
            else:
                raise NotFound(f"{cls.__name__} not found")
        return res

    @classmethod
    def cache_get(cls, shard_id: str, expression: Any, raise_if_none=None) -> Union['ShardedModel', None]:
        if expression is None:
            return None
        cache_key = f"{cls.collection}.{shard_id}.{expression}"
        getter = partial(cls.get, shard_id, expression, raise_if_none)
        constructor = partial(cls, shard_id=shard_id)
        obj = cls._cache_get(cache_key, getter, constructor)
        obj.shard_id = shard_id
        return obj

    def invalidate(self, _id=None) -> None:
        if _id is None:
            _id = self._id
        cache_key_id = f"{self.collection}.{self._shard_id}.{_id}"
        self._invalidate(cache_key_id)
        if self.KEY_FIELD is not None and self.KEY_FIELD != "_id":
            cache_key_keyfield = f"{self.collection}.{self._shard_id}.{getattr(self, self.KEY_FIELD)}"
            self._invalidate(cache_key_keyfield)

    @classmethod
    def destroy_all(cls, shard_id: str) -> None:
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        db.get_shard(shard_id).delete_query(
            cls.collection, cls._preprocess_query({}))

    @classmethod
    def destroy_many(cls, shard_id: str, query: Dict) -> None:
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        db.get_shard(shard_id).delete_query(
            cls.collection, cls._preprocess_query(query))

    @classmethod
    def update_many(cls, shard_id: str, query: Dict, attrs: Dict) -> None:
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        db.get_shard(shard_id).update_query(
            cls.collection, cls._preprocess_query(query), attrs)

    def __repr__(self) -> str:
        attributes = [f"_shard_id=%r" % self._shard_id]
        attributes += ["%s=%r" % (a, getattr(self, a))
                      for a in self.__fields__]
        return '%s(\n    %s\n)' % (self.__class__.__name__, ',\n    '.join(attributes))



import socket
import random
from time import time, sleep

from ..context import ctx

from .base_queue import BaseQueue, DEFAULT_ACK_TIMEOUT, DEFAULT_RETRIES
from .task import BaseTask, TaskSendError
from glasskit import json


class RedisQueue(BaseQueue):
    def __init__(self, qcfg):
        super(RedisQueue, self).__init__(qcfg)

        try:
            from redis import Redis
        except ImportError:
            raise RuntimeError(
                "RedisQueue is not available, redis drivers are not installed"
            )

        self.ack_timeout = qcfg.get("ack_timeout", DEFAULT_ACK_TIMEOUT)
        self.retries = qcfg.get("retries", DEFAULT_RETRIES)

        # client instances
        self._conn = None
        self._ackconn = None
        self._ps = None
        self.prefix = self.cfg.get("channel", "ueq")

        fqdn = socket.gethostname()
        rand = random.randint(0, 10000)
        self.msgchannel = f"{self.prefix}:{fqdn}:{rand}"
        self.ackchannel = f"{self.prefix}:{fqdn}:{rand}:ack"

    def init_conn(self):
        from redis import Redis

        ctx.log.info("creating a new redis connection")
        host = self.cfg.get("host", "127.0.0.1")
        port = self.cfg.get("port", 6379)
        db = self.cfg.get("dbname", 0)
        password = self.cfg.get("password")
        r = Redis(host=host, port=port, db=db, password=password)
        return r

    @property
    def conn(self):
        if self._conn is None:
            self._conn = self.init_conn()
        return self._conn

    @property
    def ackconn(self):
        if self._ackconn is None:
            self._ackconn = self.init_conn()
        return self._ackconn

    @property
    def ps(self):
        if self._ps is None:
            self._ps = self.conn.pubsub(ignore_subscribe_messages=True)
        return self._ps

    @staticmethod
    def wait_for_msg(pubsub, timeout):
        cancel_at = time() + timeout
        while time() < cancel_at:
            msg = pubsub.get_message()
            if msg:
                return msg
            sleep(0.01)
        return None

    def _enqueue(self, task):
        ack = None
        ackps = self.ackconn.pubsub(ignore_subscribe_messages=True)
        retries = self.retries
        while retries > 0:
            chan, ackchan = self.get_random_channel()
            if chan is None:
                retries -= 1
                if retries > 0:
                    sleep(0.5)
                    ctx.log.debug(
                        "no active channels found, resending, %d retries left", retries
                    )
                    continue
                else:
                    raise TaskSendError("no active channels")
            ackps.subscribe(ackchan)
            dump = json.dumps(task.to_message())
            self.conn.publish(chan, dump)
            ack = self.wait_for_msg(ackps, self.ack_timeout)
            ackps.unsubscribe(ackchan)
            if ack:
                break
            retries -= 1
            if retries > 0:
                ctx.log.debug(
                    "error receiving ack for task id %s, resending, %d retries left",
                    task.id,
                    retries,
                )
            else:
                raise TaskSendError(f"error receiving ack after {self.retries} retries")

        if ack:
            recvchan = ack["channel"].decode()
            receiver = recvchan[len(self.prefix) + 1 : -len(self.ACK_POSTFIX)]
            task.set_recv_by(receiver)

        return ack

    def ack(self, task_id):
        self.conn.publish(self.ackchannel, task_id)

    def subscribe(self):
        return self.ps.subscribe(self.msgchannel)

    def get_message(self, **kwrds):
        return self.ps.get_message(**kwrds)

    @property
    def tasks(self):
        self.subscribe()
        for redismsg in self.ps.listen():
            try:
                try:
                    msg = json.loads(redismsg["data"])
                except:
                    ctx.log.error("malformed message dump: %s", redismsg)
                    continue
                task = BaseTask.from_message(msg)
                self.ack(task.id)
                task.set_recv_by(self.msgchannel[len(self.prefix) + 1 :])
                yield task
            except Exception as e:
                ctx.log.error("error receiving message: %s", e)

    def list_active_channels(self):
        # the only way to get all the channels
        # active on server from redis client is
        # running an explicit PUBSUB channels command
        channels = [ch.decode() for ch in self.conn.execute_command("PUBSUB channels")]
        channels = [
            ch
            for ch in channels
            if ch.startswith(self.prefix) and not ch.endswith(self.ACK_POSTFIX)
        ]
        return channels

    def get_random_channel(self):
        channels = self.list_active_channels()
        if len(channels) == 0:
            return None, None
        rand = random.randrange(0, len(channels))
        return channels[rand], channels[rand] + self.ACK_POSTFIX



from uuid import uuid4
from datetime import datetime, timedelta
from flask.sessions import SessionInterface, SessionMixin
from random import random
from werkzeug.datastructures import CallbackDict

from . import ctx


class MongoSession(CallbackDict, SessionMixin):
    def __init__(self, initial=None, sid=None):
        CallbackDict.__init__(self, initial)
        self.sid = sid
        self.modified = False


class MongoSessionInterface(SessionInterface):
    def __init__(self, collection_name='sessions'):
        self.collection_name = collection_name

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if sid:
            stored_session = ctx.db.meta.get_session(sid, collection=self.collection_name)
            if stored_session:
                if stored_session.get('expiration') > datetime.utcnow():
                    return MongoSession(initial=stored_session['data'], sid=stored_session['sid'])
        else:
            sid = str(uuid4())
        return MongoSession(sid=sid)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)

        if not session:
            response.delete_cookie(app.session_cookie_name, domain=domain)
            return

        if session.modified:
            session.permanent = True
            expiration = self.get_expiration_time(app, session)
            if not expiration:
                expiration = datetime.now() + timedelta(hours=1)
            ctx.db.meta.update_session(session.sid, session, expiration, collection=self.collection_name)
            response.set_cookie(app.session_cookie_name, session.sid,
                                expires=self.get_expiration_time(app, session),
                                httponly=True, domain=domain)

        if ctx.cfg.get("session_auto_cleanup", True):
            if random() < ctx.cfg.get("session_auto_cleanup_trigger", 0.05):
                ctx.log.info("Cleaning up sessions")
                ctx.db.meta.cleanup_sessions()



import os
from flask import has_request_context, g
from datetime import datetime
from bson.objectid import ObjectId, InvalidId
from urllib.parse import urlencode, unquote, urlparse, parse_qsl, ParseResult
from uuid import uuid4
from . import json


def get_py_files(directory):
    return [f for f in os.listdir(directory) if f.endswith(".py")]


def get_modules(directory):
    return [x[:-3] for x in get_py_files(directory) if x != "__init__.py"]


def resolve_id(id_):
    # ObjectId(None) apparently generates a new unique object id
    # which is not a behaviour we need
    if id_ is not None:
        try:
            objid_expr = ObjectId(id_)
            if str(objid_expr) == id_:
                id_ = objid_expr
        except (InvalidId, TypeError):
            pass
    return id_


def cursor_to_list(crs, fields=None):
    return [x.to_dict(fields) for x in crs]


def clear_aux_fields(data):
    return {k: v for k, v in data.items() if not k.startswith("_")}


def uuid4_string():
    return str(uuid4())


def get_user_from_app_context():
    if not has_request_context():
        return None
    return g.user


# Mongo stores datetime rounded to milliseconds as it's date abilities are powered by V8
# Following is useful to avoid inconsistencies in unit tests
def now():
    dt = datetime.utcnow()
    dt = dt.replace(microsecond=dt.microsecond // 1000 * 1000)
    return dt


NilObjectId = ObjectId("000000000000000000000000")


def check_dicts_are_equal(dict1, dict2):
    """deep dict compare helper"""
    if dict1 == dict2:
        # the same object
        return True

    if len(dict1) != len(dict2):
        return False

    for k, v in dict1.items():
        if k not in dict2:
            return False
        if v == dict2[k]:
            continue
        if type(v) != type(dict2[k]):
            return False
        if type(v) is dict and check_dicts_are_equal(v, dict2[k]):
            continue
        if type(v) is list and check_lists_are_equal(v, dict2[k]):
            continue
        return False

    return True


def check_lists_are_equal(list1, list2):
    """deep list compare helper"""
    if list1 == list2:
        # the same object
        return True

    if len(list1) != len(list2):
        return False

    for i, e in enumerate(list1):
        if e == list2[i]:
            continue
        if type(e) != type(list2[i]):
            return False
        if type(e) is dict and check_dicts_are_equal(e, list2[i]):
            continue
        if type(e) is list and check_lists_are_equal(e, list2[i]):
            continue
        return False


def add_url_params(url, params):
    """ Add GET params to provided URL being aware of existing.

    :param url: string of target URL
    :param params: dict containing requested params to be added
    :return: string with updated URL

    >> url = 'http://stackoverflow.com/test?answers=true'
    >> new_params = {'answers': False, 'data': ['some','values']}
    >> add_url_params(url, new_params)
    'http://stackoverflow.com/test?data=some&data=values&answers=false'

    Taken from https://stackoverflow.com/questions/2506379/add-params-to-given-url-in-python
    Thanks to Sapphire64
    """

    url = unquote(url)
    parsed_url = urlparse(url)
    get_args = parsed_url.query
    parsed_get_args = dict(parse_qsl(get_args))
    parsed_get_args.update(params)
    parsed_get_args.update(
        {
            k: json.dumps(v)
            for k, v in parsed_get_args.items()
            if isinstance(v, (bool, dict))
        }
    )
    encoded_get_args = urlencode(parsed_get_args, doseq=True)
    new_url = ParseResult(
        parsed_url.scheme,
        parsed_url.netloc,
        parsed_url.path,
        parsed_url.params,
        encoded_get_args,
        parsed_url.fragment,
    ).geturl()
    return new_url
